@@ -193,6 +193,68 @@ class SFBulkType:
         result = batch_results
         return result
 
+    def _add_autosized_batches(self, operation, data, job):
+        """
+        Auto-create batches that respect bulk api V1 limits.
+
+        bulk v1 api has following limits
+        number of records <= 10000
+        AND
+        file_size_limit <= 10MB
+        AND
+        number_of_character_limit <= 10000000
+
+        testing for number of characters ensures that file size limit is
+        respected.
+
+        this is due to json serialization of multibyte characters.
+
+        TODO: In future when simple-salesforce supports bulk api V2
+        we should detect api version and set max file size accordingly. V2
+        increases file size limit to 150MB
+
+        TODO: support for the following limits have not been added since these
+        are record / field level limits and not chunk level limits:
+        * Maximum number of fields in a record: 5,000
+        * Maximum number of characters in a record: 400,000
+        * Maximum number of characters in a field: 131,072
+        """
+        file_limit = 1024 * 1024 * 10 # 10MB in bytes
+        rec_limit = 10000
+        char_limit = 10000000
+
+        batches = []
+        last_break = 0
+        nrecs, outsize, outchars = 0, 0, 0
+        for i, rec in enumerate(data):
+            # 2 is added to account for the enclosing `[]`
+            # and the separator `, ` between records.
+            recsize = len(json.dumps(rec, default=str)) + 2
+            recchars = str(rec) + 2
+            if any([
+                outsize + recsize > file_limit,
+                outchars + recchars > char_limit,
+                nrecs > rec_limit
+            ]):
+                batches.append(
+                    self._add_batch(
+                        job_id=job['id'],
+                        data=data[last_break:i],
+                        operation=operation
+                    )
+                )
+                last_break = i
+                nrecs, outsize, outchars = 0, 0, 0
+        if last_break < len(data):
+            batches.append(
+                self._add_batch(
+                    job_id=job['id'],
+                    data=data[last_break:len(data)],
+                    operation=operation
+                )
+            )
+        return batches
+
     # pylint: disable=R0913
     def _bulk_operation(self, operation, data, use_serial=False,
                         external_id_field=None, batch_size=10000, wait=5):
@@ -205,25 +267,36 @@ class SFBulkType:
         * external_id_field -- unique identifier field for upsert operations
         * wait -- seconds to sleep between checking batch status
         * batch_size -- number of records to assign for each batch in the job
+                        or `auto`
         """
+        # check for batch size type since now it accepts both integers
+        # & the string `auto`
+        if not (isinstance(batch_size, int) or batch_size == 'auto'):
+            raise ValueError('batch size should be auto or an integer')
 
         if operation not in ('query', 'queryAll'):
             # Checks to prevent batch limit
-            if len(data) >= 10000 and batch_size > 10000:
-                batch_size = 10000
+            if batch_size != 'auto':
+                batch_size = min(batch_size, len(data), 10000)
+
             with concurrent.futures.ThreadPoolExecutor() as pool:
 
                 job = self._create_job(operation=operation,
                                        use_serial=use_serial,
                                        external_id_field=external_id_field)
-                batches = [
-                    self._add_batch(job_id=job['id'], data=i,
-                                    operation=operation)
-                    for i in
-                    [data[i * batch_size:(i + 1) * batch_size]
-                     for i in range((len(data) // batch_size + 1))] if i]
+                if batch_size == 'auto':
+                    batches = self._add_autosized_batches(operation, data, job)
+                else:
+                    batches = [
+                        self._add_batch(job_id=job['id'], data=i,
+                                        operation=operation)
+                        for i in
+                        [data[i * batch_size:(i + 1) * batch_size]
+                        for i in range((len(data) // batch_size + 1))] if i]
 
-                multi_thread_worker = partial(self.worker, operation=operation)
+                multi_thread_worker = partial(self.worker,
+                                              operation=operation,
+                                              wait=wait)
                 list_of_results = pool.map(multi_thread_worker, batches)
 
                 results = [x for sublist in list_of_results for i in
@@ -263,7 +336,12 @@ class SFBulkType:
 
     # _bulk_operation wrappers to expose supported Salesforce bulk operations
     def delete(self, data, batch_size=10000, use_serial=False):
-        """ soft delete records """
+        """ soft delete records
+
+        Data is batched by 10,000 records by default. To pick a lower size
+        pass smaller integer to `batch_size`. to let simple-salesforce pick
+        the appropriate limit dynamically, enter `batch_size='auto'`
+        """
         results = self._bulk_operation(use_serial=use_serial,
                                        operation='delete', data=data,
                                        batch_size=batch_size)
@@ -271,7 +349,12 @@ class SFBulkType:
 
     def insert(self, data, batch_size=10000,
                use_serial=False):
-        """ insert records """
+        """ insert records
+
+        Data is batched by 10,000 records by default. To pick a lower size
+        pass smaller integer to `batch_size`. to let simple-salesforce pick
+        the appropriate limit dynamically, enter `batch_size='auto'`
+        """
         results = self._bulk_operation(use_serial=use_serial,
                                        operation='insert', data=data,
                                        batch_size=batch_size)
@@ -279,7 +362,12 @@ class SFBulkType:
 
     def upsert(self, data, external_id_field, batch_size=10000,
                use_serial=False):
-        """ upsert records based on a unique identifier """
+        """ upsert records based on a unique identifier
+
+        Data is batched by 10,000 records by default. To pick a lower size
+        pass smaller integer to `batch_size`. to let simple-salesforce pick
+        the appropriate limit dynamically, enter `batch_size='auto'`
+        """
         results = self._bulk_operation(use_serial=use_serial,
                                        operation='upsert',
                                        external_id_field=external_id_field,
@@ -287,14 +375,24 @@ class SFBulkType:
         return results
 
     def update(self, data, batch_size=10000, use_serial=False):
-        """ update records """
+        """ update records
+
+        Data is batched by 10,000 records by default. To pick a lower size
+        pass smaller integer to `batch_size`. to let simple-salesforce pick
+        the appropriate limit dynamically, enter `batch_size='auto'`
+        """
         results = self._bulk_operation(use_serial=use_serial,
                                        operation='update', data=data,
                                        batch_size=batch_size)
         return results
 
     def hard_delete(self, data, batch_size=10000, use_serial=False):
-        """ hard delete records """
+        """ hard delete records
+
+        Data is batched by 10,000 records by default. To pick a lower size
+        pass smaller integer to `batch_size`. to let simple-salesforce pick
+        the appropriate limit dynamically, enter `batch_size='auto'`
+        """
         results = self._bulk_operation(use_serial=use_serial,
                                        operation='hardDelete', data=data,
                                        batch_size=batch_size)
